import logging
import os
import random
import socket
from typing import Callable, Optional
from urllib.parse import quote

import aio_pika
import aiormq
from aio_pika import Channel, Connection


class AsyncRabbit:

    @staticmethod
    def get_uri(host: str,
                port: int,
                login: str,
                password: str,
                virtualhost: str) -> str:
        return f"amqps://{login}:{password}@{host}:{port}/{quote(virtualhost, safe='')}"

    def __init__(self,
                 url: Optional[str] = None,
                 *,
                 host: Optional[str] = None,
                 port: Optional[int] = None,
                 login: Optional[str] = None,
                 password: Optional[str] = None,
                 consumer_tag: str = None,
                 virtualhost: str = '/',
                 ssl: bool = True) -> None:
        self.url = url
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.virtualhost = virtualhost
        self.ssl = ssl

        if self.url is None:
            self.url = AsyncRabbit.get_uri(
                host=self.host,
                port=self.port,
                login=self.login,
                password=self.password,
                virtualhost=self.virtualhost
            )

        if consumer_tag is None:
            consumer_tag = f'{socket.gethostname()}[{os.getpid()}]'

        self._consumer_tag = consumer_tag

        self._qos = 1
        self._connection: Connection = None
        self._channel: Channel = None

    async def connect(self) -> None:
        self._connection: Connection = await aio_pika.connect_robust(url=self.url)  # type: ignore
        self._channel = await self._connection.channel()  # type: ignore
        await self._channel.set_qos(self._qos)
        logging.info('Connected to Rabbit MQ')

    async def live_channel(self) -> Channel:
        if self._channel and not self._channel.is_closed:
            return self._channel
        await self.connect()
        return self._channel

    async def register_callback(self, *,
                                exchange_name: str,
                                queue_name: str,
                                callback: Callable,
                                consumer_tag: str = None) -> None:
        """
        Add new callback for queue, needs to declare queue & add it to target exchange
        """
        channel = await self.live_channel()
        exchange = await channel.get_exchange(exchange_name)
        queue = await channel.declare_queue(queue_name, durable=True)

        await queue.bind(exchange)

        if consumer_tag is None:
            consumer_tag = self._consumer_tag
        try:
            await queue.consume(callback, consumer_tag=consumer_tag)
        except aio_pika.exceptions.DuplicateConsumerTag:
            consumer_tag = f"{self._consumer_tag}-{random.randint(0, 10000):05d}"
            logging.debug(f"Consumer_tag tag duplicated, add suffix now, new: {consumer_tag}")
            await self.register_callback(exchange_name=exchange_name,
                                         queue_name=queue_name,
                                         callback=callback,
                                         consumer_tag=consumer_tag)

    async def publish(self,
                      exchange_name: str,
                      message: aio_pika.Message,
                      routing_key: str = '') -> Optional[aiormq.types.ConfirmationFrameType]:
        channel = await self.live_channel()
        exchange = await channel.declare_exchange(exchange_name,
                                                  durable=True,
                                                  type=aio_pika.ExchangeType.FANOUT)
        return await exchange.publish(
            message=message,
            routing_key=routing_key,
        )

    async def close(self) -> None:
        """
        Close Rabbit MQ connections
        """
        await self._channel.close()
        await self._connection.close()

    async def cancel(self) -> None:
        """
        Cancel consumer receive new message
        """
        await self._channel.channel.basic_cancel(self._consumer_tag)


__all__ = [
    'AsyncRabbit'
]



import enum
import functools
import logging
import os
import random
import socket
import string
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Callable

import elasticapm
import pika
from pika.channel import Channel

logger = logging.getLogger("yodo1.rabbitmq")


class MQAction(enum.Enum):
    ack = "ack"
    nack = "nack"


class CallbackResult:
    def __init__(self, action: MQAction, *, requeue: bool = False):
        self.action = action
        self.requeue = requeue


def demo_callback(
    method_frame: pika.spec.Basic.Deliver,
    header_frame: pika.spec.BasicProperties,
    message_body: bytes,
) -> CallbackResult:
    """
    Sample callback function
    :param method_frame: method_frame from MQ Message
    :param header_frame: header_frame from MQ Message
    :param message_body: MQ Message body
    :return: whether should ack
    """
    if random.random() > 0.8:
        return CallbackResult(MQAction.ack)
    else:
        return CallbackResult(MQAction.nack)


class MultiThreadConsumer:
    def __init__(
        self,
        uri: str,
        *,
        qos: int = 10,
        max_worker: int = 10,
        apm_client: elasticapm.Client = None,
        verbose: bool = False,
    ) -> None:
        """
        :param uri: MQ URI
        :param qos: qos
        :param max_worker: thread max worker, if you want to process MQ in single thread, set it to 1.
        :param verbose
        """
        params = pika.URLParameters(uri)
        self.connection = pika.BlockingConnection(params)
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=qos)
        self.thread_pool = ThreadPoolExecutor(max_workers=max_worker)
        self.apm_client = apm_client
        self.verbose = verbose
        if self.apm_client:
            elasticapm.instrument()

    def setup_queue_consumer(
        self,
        queue_name: str,
        *,
        handler_function: Callable,
        exchange_name: str = None,
        consumer_tag: str = None,
    ) -> None:
        """
        Setup queue's callback function
        :param queue_name: target queue name
        :param handler_function: target callback function
        :param exchange_name: optional, exchange that needs to blind to the queue
        :param consumer_tag: optional, human-readable tag. default value is `{host_name}-{pid}-{random-string}`
        :return: None
        """
        self.channel.queue_declare(queue_name, durable=True)
        if exchange_name is not None:
            self.channel.queue_bind(queue_name, exchange=exchange_name, routing_key="")

        if consumer_tag is None:
            random_id = "".join(
                random.choices(string.ascii_lowercase + string.digits, k=6)
            )
            consumer_tag = f"{socket.gethostname()}[{os.getpid()}]-{random_id}"

        queue_thread_handler = functools.partial(
            self._handle_message,
            handler_function=handler_function,
            _queue_name=queue_name,
        )
        self.channel.basic_consume(
            queue_name, queue_thread_handler, consumer_tag=consumer_tag
        )

    def start_consuming(self) -> None:
        """
        Start consuming, will block the main thread
        """
        logger.info("Start consuming.")
        self.channel.start_consuming()

    def stop_consuming(self) -> None:
        """
        Stop consuming
        """
        self.channel.stop_consuming()
        logger.info("Stop consuming.")

    def close(self) -> None:
        """
        Close the consumer and handle remaining received messages.
        """
        logger.info("Consumer Closing... Please wait until all messages consumed.")
        self.thread_pool.shutdown(wait=True)
        self.connection.close()
        logger.info("Consumer Closed.")

    def _run_message_process(
        self,
        method_frame: pika.spec.Basic.Deliver,
        header_frame: pika.spec.BasicProperties,
        message_body: bytes,
        handler_function: Callable,
        _queue_name: str,
    ) -> CallbackResult:
        if self.apm_client:
            if header_frame.headers and "traceparent" in header_frame.headers:
                parent = elasticapm.trace_parent_from_string(
                    header_frame.headers["traceparent"]
                )
                self.apm_client.begin_transaction("RabbitMQ", trace_parent=parent)

        callback_result: CallbackResult = handler_function(
            method_frame=method_frame,
            header_frame=header_frame,
            message_body=message_body,
        )

        trace_id = elasticapm.get_trace_id()
        logger.debug(
            f"{callback_result.action.value.title()} message on Queue<{_queue_name}> with "
            f"delivery_tag: {method_frame.delivery_tag} trace_id: {trace_id}"
        )

        if trace_id:
            tran_result = (
                "success" if callback_result.action == MQAction.ack else "failure"
            )
            self.apm_client.end_transaction(
                name=f"MQ QUEUE {_queue_name}", result=tran_result
            )

        return callback_result

    def _handle_ack(
        self,
        future: Future,
        channel: Channel,
        method_frame: pika.spec.Basic.Deliver,
        _queue_name: str,
    ) -> None:
        result = future.result()
        if not isinstance(result, CallbackResult):
            raise ValueError(
                "Consumer's callback function must return a CallbackResult object."
            )
        if result.action == MQAction.ack:
            channel.basic_ack(delivery_tag=method_frame.delivery_tag)

        else:
            channel.basic_nack(delivery_tag=method_frame.delivery_tag, requeue=False)

    def _handle_message(
        self,
        channel: Channel,
        method_frame: pika.spec.Basic.Deliver,
        header_frame: pika.spec.BasicProperties,
        message_body: bytes,
        handler_function: Callable,
        _queue_name: str,
    ) -> None:
        if self.verbose:
            logger.debug(
                f"received message with tag {method_frame.delivery_tag} body: {message_body.decode()}"
            )
        future = self.thread_pool.submit(
            self._run_message_process,
            method_frame=method_frame,
            header_frame=header_frame,
            message_body=message_body,
            handler_function=handler_function,
            _queue_name=_queue_name,
        )

        # Important, add_callback_threadsafe will make sure ack event run on the same thread with the channel.
        handle_ack_callback = functools.partial(
            self._handle_ack,
            future=future,
            channel=channel,
            method_frame=method_frame,
            _queue_name=_queue_name,
        )
        future.add_done_callback(
            lambda x: self.connection.add_callback_threadsafe(handle_ack_callback)
        )


if __name__ == "__main__":
    pass



import datetime

from pydantic import BaseModel


class BaseSchema(BaseModel):
    class Config:
        orm_mode = True
        json_encoders = {
            datetime.datetime: lambda dt: dt.strftime('%Y-%m-%d %H:%M:%SZ'),
            datetime.date: lambda dt: dt.strftime('%Y-%m-%d 00:00:00Z')
        }


class BaseDateSchema(BaseSchema):
    created_at: datetime.datetime
    updated_at: datetime.datetime


__all__ = [
    'BaseSchema',
    'BaseDateSchema'
]



import datetime
from typing import Dict, Any, List, TypeVar, Type, Generator

from sqlalchemy import Column, DateTime
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql import func

T = TypeVar("T")
Base = declarative_base()  # type: ignore


@compiles(CreateTable)
def _compile_create_table(element: Any, compiler: Any, **kwargs: Dict) -> str:
    """
    Change the table order, put created_at and updated_at at last.
    Put 3 columns from base model class (is_deleted, created_at, updated_at) at the end of the table.
    """
    element.columns = element.columns[2:] + element.columns[:2]
    return compiler.visit_create_table(element)


class BaseDBModel(Base):  # type: ignore
    __abstract__ = True

    created_at = Column(DateTime,
                        default=datetime.datetime.utcnow,
                        nullable=False)
    updated_at = Column(DateTime,
                        default=datetime.datetime.utcnow,
                        onupdate=datetime.datetime.utcnow,
                        server_onupdate=func.now(),
                        nullable=False)

    def update(self, data: Dict) -> None:
        for key, value in data.items():
            setattr(self, key, value)

    def is_saved(self) -> bool:
        return self.created_at is not None

    def to_dict(self, target_cols: List[str] = None) -> Dict[str, Any]:
        if target_cols is None:
            target_cols = [c.name for c in self.__table__.columns]
        return {
            col: getattr(self, col, None)
            for col in target_cols
        }

    @classmethod
    def instance(cls: Type[T], session: Session, **kwargs) -> T:  # type: ignore  # noqa: F821
        """
        Get instance from db or create a new one
        """
        model = session.query(cls).filter_by(**kwargs).first()
        if not model:
            model = cls(**kwargs)  # type: ignore
        return model

    @classmethod
    def time_now(cls) -> datetime.datetime:
        return datetime.datetime.utcnow()

    @classmethod
    def time_delay(cls, seconds: int) -> datetime.datetime:
        return datetime.datetime.utcnow() + datetime.timedelta(seconds=seconds)


class DBManager:
    def __init__(self, engine: Any):
        self.engine = engine
        self._SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

    def get_session(self) -> Generator[None, Session, None]:
        db = None
        try:
            db = self.SessionLocal()
            yield db
        finally:
            if db:
                db.close()

    def SessionLocal(self) -> Session:
        return self._SessionLocal()

    def session(self) -> Session:
        return self.SessionLocal()


__all__ = [
    'Base',
    'BaseDBModel',
    'DBManager'
]



import json
import logging
from typing import Dict, Optional, Any
from urllib.parse import quote_plus, urlparse

import elasticapm
import httpx
from tenacity import AsyncRetrying, Retrying, stop_after_attempt, wait_random

logger = logging.getLogger("yodo1.rabbitmq")


class MQSendFailedException(Exception):
    pass


class MQExchangeDeclareFailedException(Exception):
    pass


class RabbitHttpSender:
    def __init__(
        self,
        uri: str,
        *,
        global_max_retry: int = 3,
        apm_client: elasticapm.Client = None,
        async_httpx_client: httpx.AsyncClient = httpx.AsyncClient(),
        sync_httpx_client: httpx.Client = httpx.Client(),
    ):
        """
        :param uri: rabbitmq uri
        :param global_max_retry: max retry count
        :param apm_client: elasticapm client
        :param async_httpx_client: async httpx client
        :param sync_httpx_client: sync httpx client
        """
        self.uri = uri

        uri_obj = urlparse(uri)
        self.scheme = uri_obj.scheme
        self.hostname = uri_obj.hostname
        self.username = uri_obj.username
        self.password = uri_obj.password
        self.virtual_host = uri_obj.path[1:]
        self.global_max_retry = global_max_retry
        self.apm_client: elasticapm.Client = apm_client

        if 'amqp' in self.scheme:
            logger.warning("Please use http/https for rabbitmq HTTP sender.")
            self.scheme = "https" if self.scheme == 'amqps' else "http"

        self.host = uri_obj.hostname
        if uri_obj.port:
            self.host = f'{uri_obj.hostname}:{uri_obj.port}'

        self.async_httpx_client = async_httpx_client
        self.sync_httpx_client = sync_httpx_client

        if self.apm_client:
            elasticapm.instrument()

    def declare_exchange(
        self,
        exchange_name: str,
        exchange_type: str = "fanout",
        arguments: Optional[Dict[Any, Any]] = None,
    ) -> None:
        """
        Declare exchange on start up. We should declare exchange in the code.
        We use pika here for convinience,
        :param exchange_name: str
        :param exchange_type: str
        :param arguments: Custom key/value pair arguments for the exchange
        :return:
        """
        url = f"{self.scheme}://{self.username}:{self.password}@{self.host}/api/exchanges/{quote_plus(self.virtual_host)}/{exchange_name}"

        body = {"type": exchange_type, "durable": True, "arguments": arguments or {}}
        res = self.sync_httpx_client.put(url=url, json=body)
        if res.status_code not in (201, 204):
            raise MQExchangeDeclareFailedException(res.text)
        logger.info(f"Exchange {exchange_name} declared")

    def publish(
        self,
        *,
        exchange_name: str,
        message_body: Dict,
        event_name: str = None,
        properties: Dict = None,
        routing_key: str = "",
        max_retry: int = None,
        max_delay_on_retry: int = 30,
    ) -> None:
        """
        Publish message using sync http request
        :param event_name: key event name like `app-release, user-register`
        :param exchange_name: target exchange name
        :param message_body: message body, should be able to call json.dumps
        :param properties: message properties
        :param routing_key: routing key
        :param max_retry: max retry count, will use global_max_retry when it set as empty
        :param max_delay_on_retry: max delay when retrying to send message
        :return:
        """
        # setup trace context
        traceparent_string = self._start_trace(event_name=event_name)

        if max_retry is None:
            max_retry = self.global_max_retry

        # Prepare message and url
        message = self._build_message_json(
            message_body=message_body,
            properties=properties,
            routing_key=routing_key,
            traceparent_string=traceparent_string,
            event_name=event_name,
        )
        url = f"{self.scheme}://{self.host}/api/exchanges/{quote_plus(self.virtual_host)}/{exchange_name}/publish"

        try:
            # Retry for n times before give up.
            for attempt in Retrying(
                stop=stop_after_attempt(max_retry),
                reraise=True,
                wait=wait_random(min=1, max=max_delay_on_retry),
            ):
                with attempt:
                    self._sync_publish(
                        url=url,
                        message=message,
                    )

            trace_id = elasticapm.get_trace_id()
            logger.debug(
                f"Successfully send MQ message to exchange: {exchange_name} trace_id: {trace_id}"
            )
            if self.apm_client:
                self.apm_client.end_transaction(name=f"MQ PUBLISH {event_name}", result="success")
        except Exception as e:
            logger.warning(
                "Failed to send MQ message to exchange: {exchange_name} trace_id: {trace_id}"
            )
            if self.apm_client:
                self.apm_client.capture_exception()
                self.apm_client.end_transaction(name=f"MQ PUBLISH {event_name}", result="failure")
            raise e

    async def async_publish(
        self,
        *,
        exchange_name: str,
        message_body: Dict,
        event_name: str = None,
        properties: Optional[Dict] = None,
        routing_key: str = "",
        max_retry: int = None,
        max_delay_on_retry: int = 30,
    ) -> None:
        """
        Send mq message with async function
        :param exchange_name: target exchange name
        :param message_body: message body, should be able to call json.dumps
        :param properties: message properties
        :param routing_key: routing key
        :param max_retry: max retry count, will use global_max_retry when it set as empty
        :param max_delay_on_retry: max delay when retrying to send message
        :return:
        """
        # setup trace context
        traceparent_string = self._start_trace(event_name=event_name)

        if max_retry is None:
            max_retry = self.global_max_retry
        message = self._build_message_json(
            message_body=message_body,
            properties=properties,
            routing_key=routing_key,
            traceparent_string=traceparent_string,
            event_name=event_name,
        )
        url = f"{self.scheme}://{self.host}/api/exchanges/{quote_plus(self.virtual_host)}/{exchange_name}/publish"
        try:
            # Retry for n times before give up.
            for attempt in AsyncRetrying(
                stop=stop_after_attempt(max_retry),
                reraise=True,
                wait=wait_random(min=1, max=max_delay_on_retry),
            ):
                with attempt:
                    await self._async_publish(
                        url=url,
                        message=message,
                    )
            trace_id = elasticapm.get_trace_id()
            logger.debug(
                f"Successfully send MQ message to exchange: {exchange_name} trace_id: {trace_id}"
            )
            # If successly send message
            if self.apm_client:
                self.apm_client.end_transaction(name=f"MQ PUBLISH {event_name}", result="success")
        except Exception as e:
            logger.warning(
                "Failed to send MQ message to exchange: {exchange_name} trace_id: {trace_id}"
            )
            if self.apm_client:
                self.apm_client.capture_exception()
                self.apm_client.end_transaction(name=f"MQ PUBLISH {event_name}", result="failure")
            raise e

    async def close(self) -> None:
        self.sync_httpx_client.close()
        await self.async_httpx_client.aclose()

    @staticmethod
    def _build_message_json(
        *,
        message_body: Dict,
        properties: Optional[Dict] = None,
        routing_key: str = "",
        traceparent_string: str = None,
        event_name: str = None,
    ) -> Dict:
        if properties is None:
            properties = {}

        # Add traceparent_string to headers
        if traceparent_string is not None and event_name is not None:
            if "headers" in properties:
                properties["headers"]["traceparent"] = traceparent_string
                properties["headers"]["event_name"] = event_name
            else:
                properties["headers"] = {
                    "traceparent": traceparent_string,
                    "event_name": event_name,
                }

        message = {
            "properties": properties,
            "routing_key": routing_key,
            "payload_encoding": "string",
            "payload": json.dumps(message_body),
        }
        return message

    @staticmethod
    def _check_mq_response(response: httpx.Response) -> None:
        if response.status_code == 200:
            pass
        else:
            raise MQSendFailedException(response.text)

    def _sync_publish(self, *, url: str, message: Dict) -> None:
        """
        publish MQ using sync http request
        """
        try:
            r = self.sync_httpx_client.post(
                url,
                json=message,
                auth=httpx.BasicAuth(self.username, self.password),
            )
            self._check_mq_response(r)
        except MQSendFailedException as e:
            raise e
        except Exception as e:
            raise MQSendFailedException(str(e))

    async def _async_publish(self, *, url: str, message: Dict) -> None:
        """
        publish MQ using async http request
        """
        try:
            r = await self.async_httpx_client.post(
                url,
                json=message,
                auth=httpx.BasicAuth(self.username, self.password),
            )
            self._check_mq_response(r)
        except MQSendFailedException as e:
            raise e
        except Exception as e:
            raise MQSendFailedException(str(e))

    def _start_trace(self, event_name: str) -> Optional[str]:
        if self.apm_client:
            if event_name is None:
                raise ValueError("Must set event name when using with apm client")
            # Get trace parent from the context if exists
            # This will correlate this MQ with the trigger request
            traceparent_string = elasticapm.get_trace_parent_header()
            if traceparent_string:
                parent = elasticapm.trace_parent_from_string(traceparent_string)
            else:
                parent = None
            self.apm_client.begin_transaction(
                transaction_type="RabbitMQ", trace_parent=parent
            )
            traceparent_string = elasticapm.get_trace_parent_header()
        else:
            traceparent_string = None
        return traceparent_string



import logging
from datetime import datetime, timedelta
from typing import Optional, List

import jwt
import requests
from fastapi import HTTPException, Security, Request
from fastapi.openapi.models import HTTPBearer as HTTPBearerModel
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.security.utils import get_authorization_scheme_param
from pydantic import BaseModel, Field
from starlette.status import HTTP_401_UNAUTHORIZED
import elasticapm


class HTTPBearerWithCookie(HTTPBearer):
    def __init__(
        self,
        *,
        bearerFormat: Optional[str] = None,
        scheme_name: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model = HTTPBearerModel(bearerFormat=bearerFormat)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(
        self, request: Request
    ) -> Optional[HTTPAuthorizationCredentials]:
        authorization_in_request: str = request.headers.get("Authorization")
        authorization_in_cookie: str = request.cookies.get("Authorization")
        if authorization_in_request:
            scheme, credentials = get_authorization_scheme_param(authorization_in_request)
        elif authorization_in_cookie:
            scheme, credentials = get_authorization_scheme_param(authorization_in_cookie)
        else:
            scheme, credentials = None, None
        if not (scheme and credentials):
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        if scheme.lower() != "bearer":
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED,
                    detail="Invalid authentication credentials",
                )
            else:
                return None
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)


def _exp_default_factory(expire_hours: int = 168) -> datetime:
    return datetime.utcnow() + timedelta(hours=expire_hours)


class JWTPayload(BaseModel):
    sub: str
    email: str
    name: str
    scope: List[str]
    exp: datetime = Field(default_factory=_exp_default_factory)
    iat: datetime = Field(default_factory=datetime.utcnow)


class JWTHelper:
    def __init__(self, ttl_time: int = 3600) -> None:
        self.public_key: Optional[str] = None
        self.private_key: Optional[str] = None
        self.public_key_url: Optional[str] = None
        self.scope: Optional[str] = None
        self._ttl_timedelta: timedelta = timedelta(seconds=ttl_time)
        self._last_updated_at: Optional[datetime] = None

    def _fetch_public_key(self, url: str) -> str:
        now = datetime.now()
        if self.public_key and self._is_not_expired(now):
            return self.public_key
        try:
            response = requests.get(url)
            assert response.text.startswith('-----BEGIN PUBLIC KEY-----')
            self.public_key = response.text
            self._last_updated_at = now
        except (Exception,):
            logging.error("Update public key from sso server failed.")

        if self.public_key is None:
            raise ValueError("No initial public key exists")

        return self.public_key

    def _is_not_expired(self, now: datetime) -> bool:
        if self._last_updated_at:
            return now - self._last_updated_at < self._ttl_timedelta
        else:
            return False

    def setup_with_sso_server(self, url: str, scope: Optional[str] = None) -> None:
        self.public_key = self._fetch_public_key(url=url)
        self.public_key_url = url
        self.scope = scope

    def setup_keys(self,
                   public_key: str,
                   private_key: Optional[str] = None) -> None:
        self.public_key = public_key
        self.private_key = private_key

    def encode_token(self, payload: JWTPayload) -> str:
        if self.private_key is None:
            raise ValueError('Need to setup `private_key` before call `encode_token`')
        token = jwt.encode(
            payload.dict(),
            self.private_key,
            algorithm='RS256'
        )

        # in macos, jwt.encode return bytes instead of string, so this is an ugly patch
        if isinstance(token, bytes):
            token_str = token.decode()
        else:
            token_str = token

        return token_str

    def decode_token(self, token: str) -> JWTPayload:
        if self.public_key_url:
            self.public_key = self._fetch_public_key(url=self.public_key_url)
        try:
            payload = jwt.decode(token, self.public_key, algorithms=['RS256'])
            return JWTPayload(**payload)
        except jwt.ExpiredSignatureError:
            raise HTTPException(status_code=401, detail='Token expired')
        except jwt.InvalidTokenError:
            raise HTTPException(status_code=401, detail='Invalid token')

    def current_payload(self,
                        credentials: HTTPAuthorizationCredentials = Security(HTTPBearerWithCookie())
                        ) -> JWTPayload:
        token = credentials.credentials
        payload = self.decode_token(token)
        if self.scope:
            if self.scope not in payload.scope:
                raise HTTPException(status_code=401, detail='Invalid scope')
        elasticapm.set_user_context(username=payload.name, email=payload.email, user_id=payload.sub)
        return payload


__all__ = [
    'JWTHelper',
    'JWTPayload',
    'HTTPBearerWithCookie',
]



import logging
import os
import sys

import pretty_errors
from colorlog import ColoredFormatter

pretty_errors.configure(
    separator_character='*',
    filename_display=pretty_errors.FILENAME_EXTENDED,
    line_number_first=True,
    display_link=True,
    lines_before=5,
    lines_after=2,
    line_color=pretty_errors.RED + '> ' + pretty_errors.default_config.line_color,
    code_color='  ' + pretty_errors.default_config.line_color,
    truncate_code=True,
    display_locals=True,
    display_trace_locals=True,
    truncate_locals=False,
)


def init_logger(level: str = None, style: str = 'simple') -> None:
    if level is None:
        level = os.getenv('LOG_LEVEL', 'DEBUG')
    change_log_level(level, style)


def get_color_formatter(style: str = 'simple') -> ColoredFormatter:
    if style == 'simple':
        color_format = "%(log_color)s%(levelname)-5s | [%(name)s:%(filename)s:%(lineno)d] " \
                       "%(message)s"
    else:
        color_format = "%(log_color)s[%(process)-2s] %(levelname)-5s | " \
                       "%(name)s:%(filename)s:%(lineno)d - %(message)s"

    color_formatter = ColoredFormatter(color_format,
                                       datefmt=None,
                                       reset=True,
                                       log_colors={
                                           'DEBUG': 'white',
                                           'INFO': 'green',
                                           'WARNING': 'purple',
                                           'ERROR': 'red',
                                           'CRITICAL': 'red,bg_white',
                                       },
                                       secondary_log_colors={},
                                       style='%')
    return color_formatter


def change_log_level(level: str, style: str = 'simple') -> None:
    print('----------------------')
    print('Logger init with level {}'.format(level))
    level = logging.getLevelName(level)

    color_formatter = get_color_formatter(style=style)
    print_handler = logging.StreamHandler(sys.stdout)
    print_handler.setFormatter(color_formatter)
    print_handler.setLevel(level)

    logging.basicConfig(level=logging.DEBUG, handlers=[print_handler])

    logging.info('logging init finished')


def change_default_log_levels() -> None:
    logging.getLogger("oss2.api").setLevel(logging.INFO)
    logging.getLogger("oss2.http").setLevel(logging.INFO)
    logging.getLogger("oss2.auth").setLevel(logging.INFO)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    logging.getLogger("filelock").setLevel(logging.WARNING)
    logging.getLogger('chardet.charsetprober').setLevel(logging.WARNING)
    logging.getLogger("apscheduler.scheduler").setLevel(logging.INFO)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    logging.getLogger("apscheduler.executors").setLevel(logging.INFO)
    logging.getLogger("elasticsearch").setLevel(logging.INFO)
    logging.getLogger("elasticapm").setLevel(logging.INFO)
    logging.getLogger("elasticapm.transport").setLevel(logging.INFO)
    logging.getLogger("elasticapm.metrics").setLevel(logging.INFO)
    logging.getLogger("elasticapm.conf").setLevel(logging.INFO)
    logging.getLogger("databases").setLevel(logging.INFO)
    logging.getLogger("aio_pika").setLevel(logging.INFO)
    logging.getLogger("pika").setLevel(logging.INFO)
    logging.getLogger("charset_normalizer").setLevel(logging.WARNING)


init_logger(style='simple')
change_default_log_levels()
logger = logging.getLogger('app')

if __name__ == "__main__":
    logging.info('info')
    logging.info('info')
    logging.warning('warning')
    logging.error('error')



import datetime

from fastapi.testclient import TestClient


def test_login_with_keys(client: TestClient) -> None:
    from example_api.main import auth
    from example_api.keys import PUBLIC_KEY, PRIVATE_KEY
    from yodo1.sso import JWTPayload

    auth.public_key = PUBLIC_KEY
    auth.private_key = PRIVATE_KEY

    user_payload = JWTPayload(**{
        "sub": "1234567890",
        "name": "John Doe",
        "scope": [],
        "email": "test@yodo1.com"
    })

    token = auth.encode_token(payload=user_payload)

    r = client.get("/secret_data", headers={
        'Authorization': f"Bearer {token}"
    })
    assert r.status_code == 200

    assert r.json()['user']['sub'] == '1234567890'

    # Test No Token
    r = client.get("/secret_data")
    assert r.status_code == 401

    # Test Fake Token
    r = client.get("/secret_data", headers={
        'Authorization': f"Bearer fake-token"
    })
    assert r.status_code == 401

    # Test Expired Token
    user_payload.exp = datetime.datetime.utcnow() - datetime.timedelta(hours=1)

    token = auth.encode_token(payload=user_payload)

    r = client.get("/secret_data", headers={
        'Authorization': f"Bearer {token}"
    })
    assert r.status_code == 401

    # Test Right Scope
    right_scope_user_payload = JWTPayload(**{
        "sub": "1234567890",
        "name": "John Doe",
        "scope": ['secret_service'],
        "email": "test@yodo1.com"
    })

    token = auth.encode_token(payload=right_scope_user_payload)
    auth.scope = 'secret_service'
    r = client.get("/secret_data", headers={
        'Authorization': f"Bearer {token}"
    })
    assert r.status_code == 200

    # Test Wrong Scope
    right_scope_user_payload = JWTPayload(**{
        "sub": "1234567890",
        "name": "John Doe",
        "scope": ['secret_service'],
        "email": "test@yodo1.com"
    })

    token = auth.encode_token(payload=right_scope_user_payload)
    auth.scope = 'very_secret_service'
    r = client.get("/secret_data", headers={
        'Authorization': f"Bearer {token}"
    })
    assert r.status_code == 401



from fastapi.testclient import TestClient

from example_api.base import db
from example_api.model import ItemModel


def test_db(client: TestClient) -> None:
    session = db.SessionLocal()
    total_count = session.query(ItemModel).count()

    model = ItemModel(
        title=f'Title {total_count + 1}'
    )
    session.add(model)
    session.commit()
    session.close()

    res = client.get('/items')
    items = res.json()
    assert len(items) > 0



import logging


class HealthCheckFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return 'GET /health_check' not in record.getMessage()


def patch_uvicorn_logger() -> None:
    """
    Need to call this at @app.on_event("startup"), otherwise overwrite won't affect
    :return:
    """
    uvicorn_logger = logging.getLogger('uvicorn.access')
    if uvicorn_logger:
        for h in uvicorn_logger.handlers:
            uvicorn_logger.removeHandler(h)
        uvicorn_logger.addFilter(HealthCheckFilter())
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(levelname)-5s | %(message)s"))
        uvicorn_logger.addHandler(handler)



import codecs
import pathlib
import setuptools
import re

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version() -> str:
    # Read version from file.
    version_file = pathlib.Path(__file__).parent.resolve() / "yodo1" / "__version__.py"
    with open(version_file, "r", encoding="utf-8") as f:
        version_file = f.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


with codecs.open('requirements.txt', 'r', 'utf8') as reader:
    install_requires = list(map(lambda x: x.strip(), reader.readlines()))

setuptools.setup(
    name="yodo1-toolkit",
    version=get_version(),
    author="Eliyar Eziz",
    author_email="eliyar@yodo1.com",
    description="A Yodo1 Python Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Yodo1Games/py-yodo1",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(exclude=('tests',)),
    install_requires=install_requires,
    python_requires=">=3.6",
)



import logging
import random
import time

logger = logging.getLogger('yodo1.progress')


class ProgressBar:
    WIDTH = 30

    def __init__(self, total: int, desc: str, step: int = 50) -> None:
        self.index: int = 0
        self.total: int = total
        self.step: int = step
        self.desc: str = desc

        self._start_at: float = 0

    def _format_time(self, secs: float) -> str:
        m, s = divmod(int(secs), 60)
        h, m = divmod(m, 60)
        if h == 0:
            return f'{m:02.0f}:{s:02.0f}'
        else:
            return f'{h:d}:{m:02.0f}:{s:02.0f}'

    def _get_time_string(self) -> str:
        # Fix crash when index == 0
        if self.index == 0:
            return ""
        spend_time = time.time() - self._start_at
        estimated_time = spend_time / self.index * self.total

        return f"{self._format_time(spend_time)}<{self._format_time(estimated_time)}"

    def update(self, n: int = 1) -> None:
        if self.index == 0:
            self._start_at = time.time()

        self.index += n

        if self.total == 0:
            return

        if int(self.index) % self.step == 0 or self.index == self.total:
            fill_count = int(ProgressBar.WIDTH * self.index / self.total)
            if fill_count > ProgressBar.WIDTH:
                fill_count = ProgressBar.WIDTH

            progress = f" {100 * self.index / self.total:.1f}%"
            text = f"{progress:<8s}|" + '#' * fill_count + '-' * (ProgressBar.WIDTH - fill_count) + "| " + f"{self.index}/{self.total} "
            text += f"[{self._get_time_string()}] {self.desc}"
            logger.info(text)


if __name__ == '__main__':
    import logging

    logging.basicConfig(level='DEBUG')
    p = ProgressBar(total=20, desc="Hacking ...", step=20 // 6)

    for i in range(20):
        p.update()
        time.sleep(random.randint(0, 60) / 300)
